@@ -17,10 +17,12 @@
  * the use of this software.
  */
 
+#include <libaudcore/audstrings.h>
 #include <libaudcore/drct.h>
 #include <libaudcore/equalizer.h>
 #include <libaudcore/interface.h>
 #include <libaudcore/playlist.h>
+#include <libaudcore/plugins.h>
 #include <libaudcore/runtime.h>
 #include <libaudcore/tuple.h>
 
@@ -36,10 +38,18 @@ typedef GDBusMethodInvocation Invoc;
 #define FINISH2(name, ...) \
  obj_audacious_complete_##name (obj, invoc, __VA_ARGS__)
 
-static int current_playlist ()
+static bool prefer_playing = true;
+
+static Playlist current_playlist ()
 {
-    int list = aud_playlist_get_playing ();
-    return (list >= 0) ? list : aud_playlist_get_active ();
+    Playlist list;
+
+    if (prefer_playing)
+        list = Playlist::playing_playlist ();
+    if (list == Playlist ())
+        list = Playlist::active_playlist ();
+
+    return list;
 }
 
 #define CURRENT current_playlist ()
@@ -55,28 +65,28 @@ static Index<PlaylistAddItem> strv_to_index (const char * const * strv)
 
 static gboolean do_add (Obj * obj, Invoc * invoc, const char * file)
 {
-    aud_playlist_entry_insert (CURRENT, -1, file, Tuple (), false);
+    CURRENT.insert_entry (-1, file, Tuple (), false);
     FINISH (add);
     return true;
 }
 
 static gboolean do_add_list (Obj * obj, Invoc * invoc, const char * const * filenames)
 {
-    aud_playlist_entry_insert_batch (CURRENT, -1, strv_to_index (filenames), false);
+    CURRENT.insert_items (-1, strv_to_index (filenames), false);
     FINISH (add_list);
     return true;
 }
 
 static gboolean do_add_url (Obj * obj, Invoc * invoc, const char * url)
 {
-    aud_playlist_entry_insert (CURRENT, -1, url, Tuple (), false);
+    CURRENT.insert_entry (-1, url, Tuple (), false);
     FINISH (add_url);
     return true;
 }
 
 static gboolean do_advance (Obj * obj, Invoc * invoc)
 {
-    aud_drct_pl_next ();
+    CURRENT.next_song (aud_get_bool (nullptr, "repeat"));
     FINISH (advance);
     return true;
 }
@@ -95,22 +105,36 @@ static gboolean do_balance (Obj * obj, Invoc * invoc)
 
 static gboolean do_clear (Obj * obj, Invoc * invoc)
 {
-    int playlist = CURRENT;
-    aud_playlist_entry_delete (playlist, 0, aud_playlist_entry_count (playlist));
+    CURRENT.remove_all_entries ();
     FINISH (clear);
     return true;
 }
 
+static gboolean do_config_get (Obj * obj, Invoc * invoc, const char * section, const char * name)
+{
+    String value = aud_get_str (section[0] ? section : nullptr, name);
+    FINISH2 (config_get, value);
+    return true;
+}
+
+static gboolean do_config_set (Obj * obj, Invoc * invoc, const char * section,
+ const char * name, const char * value)
+{
+    aud_set_str (section[0] ? section : nullptr, name, value);
+    FINISH (config_set);
+    return true;
+}
+
 static gboolean do_delete (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    aud_playlist_entry_delete (CURRENT, pos, 1);
+    CURRENT.remove_entry (pos);
     FINISH (delete);
     return true;
 }
 
 static gboolean do_delete_active_playlist (Obj * obj, Invoc * invoc)
 {
-    aud_playlist_delete (CURRENT);
+    CURRENT.remove_playlist ();
     FINISH (delete_active_playlist);
     return true;
 }
@@ -133,13 +157,13 @@ static gboolean do_equalizer_activate (Obj * obj, Invoc * invoc, gboolean active
 
 static gboolean do_get_active_playlist (Obj * obj, Invoc * invoc)
 {
-    FINISH2 (get_active_playlist, CURRENT);
+    FINISH2 (get_active_playlist, CURRENT.index ());
     return true;
 }
 
 static gboolean do_get_active_playlist_name (Obj * obj, Invoc * invoc)
 {
-    String title = aud_playlist_get_title (CURRENT);
+    String title = CURRENT.get_title ();
     FINISH2 (get_active_playlist_name, title ? title : "");
     return true;
 }
@@ -178,7 +202,7 @@ static gboolean do_get_info (Obj * obj, Invoc * invoc)
 
 static gboolean do_get_playqueue_length (Obj * obj, Invoc * invoc)
 {
-    FINISH2 (get_playqueue_length, aud_playlist_queue_count (CURRENT));
+    FINISH2 (get_playqueue_length, CURRENT.n_queued ());
     return true;
 }
 
@@ -205,14 +229,14 @@ static gboolean do_info (Obj * obj, Invoc * invoc)
 
 static gboolean do_jump (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    aud_playlist_set_position (CURRENT, pos);
+    CURRENT.set_position (pos);
     FINISH (jump);
     return true;
 }
 
 static gboolean do_length (Obj * obj, Invoc * invoc)
 {
-    FINISH2 (length, aud_playlist_entry_count (CURRENT));
+    FINISH2 (length, CURRENT.n_entries ());
     return true;
 }
 
@@ -224,9 +248,7 @@ static gboolean do_main_win_visible (Obj * obj, Invoc * invoc)
 
 static gboolean do_new_playlist (Obj * obj, Invoc * invoc)
 {
-    int playlist = CURRENT + 1;
-    aud_playlist_insert (playlist);
-    aud_playlist_set_active (playlist);
+    Playlist::insert_playlist (CURRENT.index () + 1).activate ();
     aud_drct_stop ();
     FINISH (new_playlist);
     return true;
@@ -234,7 +256,7 @@ static gboolean do_new_playlist (Obj * obj, Invoc * invoc)
 
 static gboolean do_number_of_playlists (Obj * obj, Invoc * invoc)
 {
-    FINISH2 (number_of_playlists, aud_playlist_count ());
+    FINISH2 (number_of_playlists, Playlist::n_playlists ());
     return true;
 }
 
@@ -274,7 +296,7 @@ static gboolean do_play (Obj * obj, Invoc * invoc)
 
 static gboolean do_play_active_playlist (Obj * obj, Invoc * invoc)
 {
-    aud_playlist_play (CURRENT);
+    CURRENT.start_playback ();
     FINISH (play_active_playlist);
     return true;
 }
@@ -294,7 +316,7 @@ static gboolean do_playing (Obj * obj, Invoc * invoc)
 
 static gboolean do_playlist_add (Obj * obj, Invoc * invoc, const char * list)
 {
-    aud_playlist_entry_insert (CURRENT, -1, list, Tuple (), false);
+    CURRENT.insert_entry (-1, list, Tuple (), false);
     FINISH (playlist_add);
     return true;
 }
@@ -308,60 +330,86 @@ static gboolean do_playlist_enqueue_to_temp (Obj * obj, Invoc * invoc, const cha
 
 static gboolean do_playlist_ins_url_string (Obj * obj, Invoc * invoc, const char * url, int pos)
 {
-    aud_playlist_entry_insert (CURRENT, pos, url, Tuple (), false);
+    CURRENT.insert_entry (pos, url, Tuple (), false);
     FINISH (playlist_ins_url_string);
     return true;
 }
 
 static gboolean do_playqueue_add (Obj * obj, Invoc * invoc, int pos)
 {
-    aud_playlist_queue_insert (CURRENT, -1, pos);
+    CURRENT.queue_insert (-1, pos);
     FINISH (playqueue_add);
     return true;
 }
 
 static gboolean do_playqueue_clear (Obj * obj, Invoc * invoc)
 {
-    int playlist = CURRENT;
-    aud_playlist_queue_delete (playlist, 0, aud_playlist_queue_count (playlist));
+    CURRENT.queue_remove_all ();
     FINISH (playqueue_clear);
     return true;
 }
 
 static gboolean do_playqueue_is_queued (Obj * obj, Invoc * invoc, int pos)
 {
-    bool queued = (aud_playlist_queue_find_entry (CURRENT, pos) >= 0);
+    bool queued = (CURRENT.queue_find_entry (pos) >= 0);
     FINISH2 (playqueue_is_queued, queued);
     return true;
 }
 
 static gboolean do_playqueue_remove (Obj * obj, Invoc * invoc, int pos)
 {
-    int playlist = CURRENT;
-    int qpos = aud_playlist_queue_find_entry (playlist, pos);
+    auto playlist = CURRENT;
+    int qpos = playlist.queue_find_entry (pos);
 
     if (qpos >= 0)
-        aud_playlist_queue_delete (playlist, qpos, 1);
+        playlist.queue_remove (qpos);
 
     FINISH (playqueue_remove);
     return true;
 }
 
+static gboolean do_plugin_enable (Obj * obj, Invoc * invoc, const char * name, gboolean enable)
+{
+    PluginHandle * plugin = aud_plugin_lookup_basename (name);
+    if (! plugin)
+    {
+        AUDERR ("No such plugin: %s\n", name);
+        return false;
+    }
+
+    aud_plugin_enable (plugin, enable);
+    FINISH (plugin_enable);
+    return true;
+}
+
+static gboolean do_plugin_is_enabled (Obj * obj, Invoc * invoc, const char * name)
+{
+    PluginHandle * plugin = aud_plugin_lookup_basename (name);
+    if (! plugin)
+    {
+        AUDERR ("No such plugin: %s\n", name);
+        return false;
+    }
+
+    FINISH2 (plugin_is_enabled, aud_plugin_get_enabled (plugin));
+    return true;
+}
+
 static gboolean do_position (Obj * obj, Invoc * invoc)
 {
-    FINISH2 (position, aud_playlist_get_position (CURRENT));
+    FINISH2 (position, CURRENT.get_position ());
     return true;
 }
 
 static gboolean do_queue_get_list_pos (Obj * obj, Invoc * invoc, unsigned qpos)
 {
-    FINISH2 (queue_get_list_pos, aud_playlist_queue_get_entry (CURRENT, qpos));
+    FINISH2 (queue_get_list_pos, CURRENT.queue_get_entry (qpos));
     return true;
 }
 
 static gboolean do_queue_get_queue_pos (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    FINISH2 (queue_get_queue_pos, aud_playlist_queue_find_entry (CURRENT, pos));
+    FINISH2 (queue_get_queue_pos, CURRENT.queue_find_entry (pos));
     return true;
 }
 
@@ -372,6 +420,25 @@ static gboolean do_quit (Obj * obj, Invoc * invoc)
     return true;
 }
 
+static gboolean do_record (Obj * obj, Invoc * invoc)
+{
+    if (aud_drct_get_record_enabled ())
+        aud_set_bool (nullptr, "record", ! aud_get_bool (nullptr, "record"));
+
+    FINISH (record);
+    return true;
+}
+
+static gboolean do_recording (Obj * obj, Invoc * invoc)
+{
+    bool recording = false;
+    if (aud_drct_get_record_enabled ())
+        recording = aud_get_bool (nullptr, "record");
+
+    FINISH2 (recording, recording);
+    return true;
+}
+
 static gboolean do_repeat (Obj * obj, Invoc * invoc)
 {
     FINISH2 (repeat, aud_get_bool (nullptr, "repeat"));
@@ -380,7 +447,7 @@ static gboolean do_repeat (Obj * obj, Invoc * invoc)
 
 static gboolean do_reverse (Obj * obj, Invoc * invoc)
 {
-    aud_drct_pl_prev ();
+    CURRENT.prev_song ();
     FINISH (reverse);
     return true;
 }
@@ -392,13 +459,28 @@ static gboolean do_seek (Obj * obj, Invoc * invoc, unsigned pos)
     return true;
 }
 
-static gboolean do_set_active_playlist (Obj * obj, Invoc * invoc, int playlist)
+static gboolean do_select_displayed_playlist (Obj * obj, Invoc * invoc)
 {
-    aud_playlist_set_active (playlist);
+    prefer_playing = false;
+    FINISH (select_displayed_playlist);
+    return true;
+}
 
-    // check that the requested playlist exists before switching playback
-    if (aud_playlist_get_active () == playlist && aud_drct_get_playing ())
-        aud_playlist_play (playlist, aud_drct_get_paused ());
+static gboolean do_select_playing_playlist (Obj * obj, Invoc * invoc)
+{
+    prefer_playing = true;
+    FINISH (select_playing_playlist);
+    return true;
+}
+
+static gboolean do_set_active_playlist (Obj * obj, Invoc * invoc, int index)
+{
+    auto playlist = Playlist::by_index (index);
+
+    playlist.activate ();
+
+    if (prefer_playing && aud_drct_get_playing ())
+        playlist.start_playback (aud_drct_get_paused ());
 
     FINISH (set_active_playlist);
     return true;
@@ -406,7 +488,7 @@ static gboolean do_set_active_playlist (Obj * obj, Invoc * invoc, int playlist)
 
 static gboolean do_set_active_playlist_name (Obj * obj, Invoc * invoc, const char * title)
 {
-    aud_playlist_set_title (CURRENT, title);
+    CURRENT.set_title (title);
     FINISH (set_active_playlist_name);
     return true;
 }
@@ -522,21 +604,21 @@ static gboolean do_shuffle (Obj * obj, Invoc * invoc)
 
 static gboolean do_song_filename (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    String filename = aud_playlist_entry_get_filename (CURRENT, pos);
+    String filename = CURRENT.entry_filename (pos);
     FINISH2 (song_filename, filename ? filename : "");
     return true;
 }
 
 static gboolean do_song_frames (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    Tuple tuple = aud_playlist_entry_get_tuple (CURRENT, pos);
+    Tuple tuple = CURRENT.entry_tuple (pos);
     FINISH2 (song_frames, aud::max (0, tuple.get_int (Tuple::Length)));
     return true;
 }
 
 static gboolean do_song_length (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    Tuple tuple = aud_playlist_entry_get_tuple (CURRENT, pos);
+    Tuple tuple = CURRENT.entry_tuple (pos);
     int length = aud::max (0, tuple.get_int (Tuple::Length));
     FINISH2 (song_length, length / 1000);
     return true;
@@ -544,7 +626,7 @@ static gboolean do_song_length (Obj * obj, Invoc * invoc, unsigned pos)
 
 static gboolean do_song_title (Obj * obj, Invoc * invoc, unsigned pos)
 {
-    Tuple tuple = aud_playlist_entry_get_tuple (CURRENT, pos);
+    Tuple tuple = CURRENT.entry_tuple (pos);
     String title = tuple.get_str (Tuple::FormattedTitle);
     FINISH2 (song_title, title ? title : "");
     return true;
@@ -554,35 +636,37 @@ static gboolean do_song_tuple (Obj * obj, Invoc * invoc, unsigned pos, const cha
 {
     Tuple::Field field = Tuple::field_by_name (key);
     Tuple tuple;
-    GVariant * var = nullptr;
+    GVariant * var;
 
     if (field >= 0)
-        tuple = aud_playlist_entry_get_tuple (CURRENT, pos);
+        tuple = CURRENT.entry_tuple (pos);
 
-    if (tuple)
+    switch (tuple.get_value_type (field))
     {
-        switch (tuple.get_value_type (field))
-        {
-        case Tuple::String:
-            var = g_variant_new_string (tuple.get_str (field));
-            break;
+    case Tuple::String:
+        var = g_variant_new_string (tuple.get_str (field));
+        break;
 
-        case Tuple::Int:
-            var = g_variant_new_int32 (tuple.get_int (field));
-            break;
+    case Tuple::Int:
+        var = g_variant_new_int32 (tuple.get_int (field));
+        break;
 
-        default:
-            break;
-        }
+    default:
+        var = g_variant_new_string ("");
+        break;
     }
 
-    if (! var)
-        var = g_variant_new_string ("");
-
     FINISH2 (song_tuple, g_variant_new_variant (var));
     return true;
 }
 
+static gboolean do_startup_notify (Obj * obj, Invoc * invoc, const char * id)
+{
+    aud_ui_startup_notify (id);
+    FINISH (startup_notify);
+    return true;
+}
+
 static gboolean do_status (Obj * obj, Invoc * invoc)
 {
     const char * status = "stopped";
@@ -620,28 +704,28 @@ static gboolean do_time (Obj * obj, Invoc * invoc)
 
 static gboolean do_toggle_auto_advance (Obj * obj, Invoc * invoc)
 {
-    aud_set_bool (nullptr, "no_playlist_advance", ! aud_get_bool (nullptr, "no_playlist_advance"));
+    aud_toggle_bool (nullptr, "no_playlist_advance");
     FINISH (toggle_auto_advance);
     return true;
 }
 
 static gboolean do_toggle_repeat (Obj * obj, Invoc * invoc)
 {
-    aud_set_bool (nullptr, "repeat", ! aud_get_bool (nullptr, "repeat"));
+    aud_toggle_bool (nullptr, "repeat");
     FINISH (toggle_repeat);
     return true;
 }
 
 static gboolean do_toggle_shuffle (Obj * obj, Invoc * invoc)
 {
-    aud_set_bool (nullptr, "shuffle", ! aud_get_bool (nullptr, "shuffle"));
+    aud_toggle_bool (nullptr, "shuffle");
     FINISH (toggle_shuffle);
     return true;
 }
 
 static gboolean do_toggle_stop_after (Obj * obj, Invoc * invoc)
 {
-    aud_set_bool (nullptr, "stop_after_current_song", ! aud_get_bool (nullptr, "stop_after_current_song"));
+    aud_toggle_bool (nullptr, "stop_after_current_song");
     FINISH (toggle_stop_after);
     return true;
 }
@@ -673,6 +757,8 @@ handlers[] =
     {"handle-auto-advance", (GCallback) do_auto_advance},
     {"handle-balance", (GCallback) do_balance},
     {"handle-clear", (GCallback) do_clear},
+    {"handle-config-get", (GCallback) do_config_get},
+    {"handle-config-set", (GCallback) do_config_set},
     {"handle-delete", (GCallback) do_delete},
     {"handle-delete-active-playlist", (GCallback) do_delete_active_playlist},
     {"handle-eject", (GCallback) do_eject},
@@ -706,13 +792,19 @@ handlers[] =
     {"handle-playqueue-clear", (GCallback) do_playqueue_clear},
     {"handle-playqueue-is-queued", (GCallback) do_playqueue_is_queued},
     {"handle-playqueue-remove", (GCallback) do_playqueue_remove},
+    {"handle-plugin-enable", (GCallback) do_plugin_enable},
+    {"handle-plugin-is-enabled", (GCallback) do_plugin_is_enabled},
     {"handle-position", (GCallback) do_position},
     {"handle-queue-get-list-pos", (GCallback) do_queue_get_list_pos},
     {"handle-queue-get-queue-pos", (GCallback) do_queue_get_queue_pos},
     {"handle-quit", (GCallback) do_quit},
+    {"handle-recording", (GCallback) do_recording},
+    {"handle-record", (GCallback) do_record},
     {"handle-repeat", (GCallback) do_repeat},
     {"handle-reverse", (GCallback) do_reverse},
     {"handle-seek", (GCallback) do_seek},
+    {"handle-select-displayed-playlist", (GCallback) do_select_displayed_playlist},
+    {"handle-select-playing-playlist", (GCallback) do_select_playing_playlist},
     {"handle-set-active-playlist", (GCallback) do_set_active_playlist},
     {"handle-set-active-playlist-name", (GCallback) do_set_active_playlist_name},
     {"handle-set-eq", (GCallback) do_set_eq},
@@ -730,6 +822,7 @@ handlers[] =
     {"handle-song-length", (GCallback) do_song_length},
     {"handle-song-title", (GCallback) do_song_title},
     {"handle-song-tuple", (GCallback) do_song_tuple},
+    {"handle-startup-notify", (GCallback) do_startup_notify},
     {"handle-status", (GCallback) do_status},
     {"handle-stop", (GCallback) do_stop},
     {"handle-stop-after", (GCallback) do_stop_after},
@@ -748,16 +841,16 @@ static unsigned owner_id = 0;
 
 static GDBusInterfaceSkeleton * skeleton = nullptr;
 
-static void name_acquired (GDBusConnection *, const char *, void *)
+static void name_acquired (GDBusConnection *, const char * name, void *)
 {
-    AUDINFO ("Owned D-Bus name (org.atheme.audacious) on session bus.\n");
+    AUDINFO ("Owned D-Bus name (%s) on session bus.\n", name);
 
     g_main_loop_quit (mainloop);
 }
 
-static void name_lost (GDBusConnection *, const char *, void *)
+static void name_lost (GDBusConnection *, const char * name, void *)
 {
-    AUDINFO ("Owning D-Bus name (org.atheme.audacious) failed, already taken?\n");
+    AUDINFO ("Owning D-Bus name (%s) failed, already taken?\n", name);
 
     g_bus_unown_name (owner_id);
     owner_id = 0;
@@ -765,8 +858,17 @@ static void name_lost (GDBusConnection *, const char *, void *)
     g_main_loop_quit (mainloop);
 }
 
+StringBuf dbus_server_name ()
+{
+    int instance = aud_get_instance ();
+    return (instance == 1) ? str_copy ("org.atheme.audacious") :
+           str_printf ("org.atheme.audacious-%d", instance);
+}
+
 StartupType dbus_server_init ()
 {
+    auto startup = StartupType::Unknown;
+
     GError * error = nullptr;
     GDBusConnection * bus = g_bus_get_sync (G_BUS_TYPE_SESSION, nullptr, & error);
     GMainContext * context;
@@ -785,7 +887,7 @@ StartupType dbus_server_init ()
     context = g_main_context_new ();
     g_main_context_push_thread_default (context);
 
-    owner_id = g_bus_own_name (G_BUS_TYPE_SESSION, "org.atheme.audacious",
+    owner_id = g_bus_own_name (G_BUS_TYPE_SESSION, dbus_server_name (),
      (GBusNameOwnerFlags) 0, nullptr, name_acquired, name_lost, nullptr, nullptr);
 
     mainloop = g_main_loop_new (context, true);
@@ -793,15 +895,14 @@ StartupType dbus_server_init ()
     g_main_loop_unref (mainloop);
     mainloop = nullptr;
 
+    if (owner_id)
+        startup = StartupType::Server;
+    else
+        startup = StartupType::Client;
+
     g_main_context_pop_thread_default (context);
     g_main_context_unref (context);
 
-    if (owner_id)
-        return StartupType::Server;
-
-    dbus_server_cleanup ();
-    return StartupType::Client;
-
 ERROR:
     if (error)
     {
@@ -809,8 +910,10 @@ ERROR:
         g_error_free (error);
     }
 
-    dbus_server_cleanup ();
-    return StartupType::Unknown;
+    if (startup != StartupType::Server)
+        dbus_server_cleanup ();
+
+    return startup;
 }
 
 void dbus_server_cleanup ()
