@@ -1,6 +1,6 @@
 /*
  * output.c
- * Copyright 2009-2013 John Lindgren
+ * Copyright 2009-2015 John Lindgren
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -25,6 +25,9 @@
 #include <string.h>
 
 #include "equalizer.h"
+#include "hook.h"
+#include "i18n.h"
+#include "interface.h"
 #include "internal.h"
 #include "plugin.h"
 #include "plugins.h"
@@ -79,10 +82,14 @@ static pthread_cond_t cond_minor = PTHREAD_COND_INITIALIZER;
 static OutputPlugin * cop; /* current (primary) output plugin */
 static OutputPlugin * sop; /* secondary output plugin */
 
+static OutputStream record_stream;
+
 static int seek_time;
 static String in_filename;
 static Tuple in_tuple;
 static int in_format, in_channels, in_rate;
+static int effect_channels, effect_rate;
+static int sec_channels, sec_rate;
 static int out_format, out_channels, out_rate;
 static int out_bytes_per_sec, out_bytes_held;
 static int64_t in_frames, out_bytes_written;
@@ -91,17 +98,32 @@ static ReplayGainInfo gain_info;
 static Index<float> buffer1;
 static Index<char> buffer2;
 
-static inline int get_format ()
+static inline int get_format (bool & automatic)
 {
+    automatic = false;
+
     switch (aud_get_int (0, "output_bit_depth"))
     {
         case 16: return FMT_S16_NE;
-        case 24: return FMT_S24_NE;
+        case 24: return FMT_S24_3NE;
         case 32: return FMT_S32_NE;
+
+        // return FMT_FLOAT for "auto" as well
+        case -1: automatic = true;
         default: return FMT_FLOAT;
     }
 }
 
+/* assumes LOCK_ALL, s_input */
+static void setup_effects ()
+{
+    effect_channels = in_channels;
+    effect_rate = in_rate;
+
+    effect_start (effect_channels, effect_rate);
+    eq_set_format (effect_channels, effect_rate);
+}
+
 /* assumes LOCK_ALL */
 static void cleanup_output ()
 {
@@ -134,74 +156,56 @@ static void cleanup_secondary ()
     sop->close_audio ();
 }
 
-/* assumes LOCK_ALL, s_output */
+/* assumes LOCK_MINOR, s_output */
 static void apply_pause ()
 {
     cop->pause (s_paused);
     vis_runner_start_stop (true, s_paused);
 }
 
-/* assumes LOCK_ALL (LOCK_MINOR for secondary),
- * s_input, valid out_channels/rate */
-static bool setup_plugin (OutputPlugin * op, int format)
-{
-    op->set_info (in_filename, in_tuple);
-    return op->open_audio (format, out_rate, out_channels);
-}
-
 /* assumes LOCK_ALL, s_input */
-static void setup_output ()
+static void setup_output (bool new_input)
 {
-    int format = get_format ();
-    int channels = in_channels;
-    int rate = in_rate;
-
-    effect_start (channels, rate);
-    eq_set_format (channels, rate);
-
-    AUDINFO ("Setup output, format %d, %d channels, %d Hz.\n", format, channels, rate);
-
-    bool keep_output = (s_output && format == out_format &&
-     channels == out_channels && rate == out_rate && ! cop->force_reopen);
-
-    /* secondary output always uses FMT_FLOAT */
-    bool keep_secondary = (s_secondary && channels == out_channels &&
-     rate == out_rate && ! sop->force_reopen);
-
-    out_format = format;
-    out_channels = channels;
-    out_rate = rate;
-
-    if (keep_output)
-        AUDINFO ("Reusing primary output stream.\n");
-    else
-    {
-        cleanup_output ();
-
-        if (cop && setup_plugin (cop, out_format))
-            s_output = true;
-        else
-        {
-            effect_flush (true);
-            cleanup_secondary ();
-            return;
-        }
-    }
-
-    if (keep_secondary)
-        AUDINFO ("Reusing secondary output stream.\n");
-    else
-    {
-        cleanup_secondary ();
-
-        if (sop && setup_plugin (sop, FMT_FLOAT))
-            s_secondary = true;
-    }
-
-    if (keep_output)
+    if (! cop)
         return;
 
-    out_bytes_per_sec = FMT_SIZEOF (format) * channels * rate;
+    bool automatic;
+    int format = get_format (automatic);
+
+    AUDINFO ("Setup output, format %d, %d channels, %d Hz.\n", format, effect_channels, effect_rate);
+
+    if (s_output && format == out_format && effect_channels == out_channels &&
+     effect_rate == out_rate && ! (new_input && cop->force_reopen))
+        return;
+
+    cleanup_output ();
+    cop->set_info (in_filename, in_tuple);
+
+    String error;
+    while (! cop->open_audio (format, effect_rate, effect_channels, error))
+    {
+        if (automatic && format == FMT_FLOAT)
+            format = FMT_S32_NE;
+        else if (automatic && format == FMT_S32_NE)
+            format = FMT_S16_NE;
+        else if (format == FMT_S24_3NE)
+            format = FMT_S24_NE; /* some output plugins support only padded 24-bit */
+        else
+        {
+            aud_ui_show_error (error ? (const char *) error : _("Error opening output stream"));
+            return;
+        }
+
+        AUDINFO ("Falling back to format %d.\n", format);
+    }
+
+    s_output = true;
+
+    out_format = format;
+    out_channels = effect_channels;
+    out_rate = effect_rate;
+
+    out_bytes_per_sec = FMT_SIZEOF (format) * out_channels * out_rate;
     out_bytes_held = 0;
     out_bytes_written = 0;
 
@@ -211,18 +215,54 @@ static void setup_output ()
         SIGNAL_MINOR;
 }
 
-/* assumes LOCK_MINOR, s_output */
-static bool flush_output (bool force)
+/* assumes LOCK_MINOR, s_input */
+static void setup_secondary (bool new_input)
 {
-    if (! effect_flush (force))
-        return false;
+    if (! sop)
+        return;
 
+    int rate, channels;
+    record_stream = (OutputStream) aud_get_int (0, "record_stream");
+
+    if (record_stream < OutputStream::AfterEffects)
+    {
+        rate = in_rate;
+        channels = in_channels;
+    }
+    else
+    {
+        rate = effect_rate;
+        channels = effect_channels;
+    }
+
+    if (s_secondary && channels == sec_channels && rate == sec_rate &&
+     ! (new_input && sop->force_reopen))
+        return;
+
+    cleanup_secondary ();
+    sop->set_info (in_filename, in_tuple);
+
+    String error;
+    if (! sop->open_audio (FMT_FLOAT, rate, channels, error))
+    {
+        aud_ui_show_error (error ? (const char *) error : _("Error recording output stream"));
+        return;
+    }
+
+    s_secondary = true;
+
+    sec_channels = channels;
+    sec_rate = rate;
+}
+
+/* assumes LOCK_MINOR, s_output */
+static void flush_output ()
+{
     out_bytes_held = 0;
     out_bytes_written = 0;
 
     cop->flush ();
     vis_runner_flush ();
-    return true;
 }
 
 static void apply_replay_gain (Index<float> & data)
@@ -236,7 +276,10 @@ static void apply_replay_gain (Index<float> & data)
     {
         float peak;
 
-        if (aud_get_bool (0, "replay_gain_album"))
+        auto mode = (ReplayGainMode) aud_get_int (0, "replay_gain_mode");
+        if ((mode == ReplayGainMode::Album) ||
+            (mode == ReplayGainMode::Automatic &&
+             (! aud_get_bool (0, "shuffle") || aud_get_bool (0, "album_shuffle"))))
         {
             factor *= powf (10, gain_info.album_gain / 20);
             peak = gain_info.album_peak;
@@ -257,24 +300,32 @@ static void apply_replay_gain (Index<float> & data)
         audio_amplify (data.begin (), 1, data.len (), & factor);
 }
 
+/* assumes LOCK_MINOR, s_secondary */
+static void write_secondary (const Index<float> & data)
+{
+    auto begin = (const char *) data.begin ();
+    auto end = (const char *) data.end ();
+
+    while (begin < end)
+        begin += sop->write_audio (begin, end - begin);
+}
+
 /* assumes LOCK_ALL, s_output */
-static void write_output_raw (Index<float> & data)
+static void write_output (Index<float> & data)
 {
     if (! data.len ())
         return;
 
+    if (s_secondary && record_stream == OutputStream::AfterEffects)
+        write_secondary (data);
+
     int out_time = aud::rescale<int64_t> (out_bytes_written, out_bytes_per_sec, 1000);
     vis_runner_pass_audio (out_time, data, out_channels, out_rate);
 
     eq_filter (data.begin (), data.len ());
 
-    if (s_secondary)
-    {
-        auto begin = (const char *) data.begin ();
-        auto end = (const char *) data.end ();
-        while (begin < end)
-            begin += sop->write_audio (begin, end - begin);
-    }
+    if (s_secondary && record_stream == OutputStream::AfterEqualizer)
+        write_secondary (data);
 
     if (aud_get_bool (0, "software_volume_control"))
     {
@@ -314,7 +365,7 @@ static void write_output_raw (Index<float> & data)
 }
 
 /* assumes LOCK_ALL, s_input, s_output */
-static bool write_output (const void * data, int size, int stop_time)
+static bool process_audio (const void * data, int size, int stop_time)
 {
     int samples = size / FMT_SIZEOF (in_format);
     bool stopped = false;
@@ -340,8 +391,15 @@ static bool write_output (const void * data, int size, int stop_time)
     else
         audio_from_int (data, in_format, buffer1.begin (), samples);
 
+    if (s_secondary && record_stream == OutputStream::AsDecoded)
+        write_secondary (buffer1);
+
     apply_replay_gain (buffer1);
-    write_output_raw (effect_process (buffer1));
+
+    if (s_secondary && record_stream == OutputStream::AfterReplayGain)
+        write_secondary (buffer1);
+
+    write_output (effect_process (buffer1));
 
     return ! stopped;
 }
@@ -350,7 +408,7 @@ static bool write_output (const void * data, int size, int stop_time)
 static void finish_effects (bool end_of_playlist)
 {
     buffer1.resize (0);
-    write_output_raw (effect_finish (buffer1, end_of_playlist));
+    write_output (effect_finish (buffer1, end_of_playlist));
 }
 
 bool output_open_audio (const String & filename, const Tuple & tuple,
@@ -364,11 +422,8 @@ bool output_open_audio (const String & filename, const Tuple & tuple,
 
     if (s_output && s_paused)
     {
-        if (! s_flushed && ! s_resetting)
-            flush_output (true);
-
-        s_paused = false;
-        apply_pause ();
+        effect_flush (true);
+        cleanup_output ();
     }
 
     s_input = true;
@@ -382,7 +437,11 @@ bool output_open_audio (const String & filename, const Tuple & tuple,
     in_rate = rate;
     in_frames = 0;
 
-    setup_output ();
+    setup_effects ();
+    setup_output (true);
+
+    if (aud_get_bool (0, "record"))
+        setup_secondary (true);
 
     UNLOCK_ALL;
     return true;
@@ -390,12 +449,12 @@ bool output_open_audio (const String & filename, const Tuple & tuple,
 
 void output_set_tuple (const Tuple & tuple)
 {
-    LOCK_ALL;
+    LOCK_MINOR;
 
     if (s_input)
         in_tuple = tuple.ref ();
 
-    UNLOCK_ALL;
+    UNLOCK_MINOR;
 }
 
 void output_set_replay_gain (const ReplayGainInfo & info)
@@ -434,7 +493,7 @@ RETRY:
             goto RETRY;
         }
 
-        good = write_output (data, size, stop_time);
+        good = process_audio (data, size, stop_time);
     }
 
     UNLOCK_ALL;
@@ -451,9 +510,13 @@ void output_flush (int time, bool force)
         {
             // allow effect plugins to prevent the flush, but
             // always flush if paused to prevent locking up
-            s_flushed = flush_output (s_paused || force);
-            if (s_paused)
-                SIGNAL_MINOR;
+            if (effect_flush (s_paused || force))
+            {
+                flush_output ();
+                s_flushed = true;
+                if (s_paused)
+                    SIGNAL_MINOR;
+            }
         }
         else
         {
@@ -577,16 +640,13 @@ static void output_reset (OutputReset type, OutputPlugin * op)
     s_resetting = true;
 
     if (s_output && ! s_flushed)
-        flush_output (true);
+        flush_output ();
 
     UNLOCK_MINOR;
     LOCK_ALL;
 
     if (type != OutputReset::EffectsOnly)
-    {
         cleanup_output ();
-        cleanup_secondary ();
-    }
 
     /* this does not reset the secondary plugin */
     if (type == OutputReset::ResetPlugin)
@@ -598,7 +658,10 @@ static void output_reset (OutputReset type, OutputPlugin * op)
         {
             /* secondary plugin may become primary */
             if (op == sop)
+            {
+                cleanup_secondary ();
                 sop = nullptr;
+            }
             else if (! op->init ())
                 op = nullptr;
         }
@@ -607,7 +670,15 @@ static void output_reset (OutputReset type, OutputPlugin * op)
     }
 
     if (s_input)
-        setup_output ();
+    {
+        if (type == OutputReset::EffectsOnly)
+            setup_effects ();
+
+        setup_output (false);
+
+        if (aud_get_bool (0, "record"))
+            setup_secondary (false);
+    }
 
     s_resetting = false;
 
@@ -683,9 +754,33 @@ bool output_plugin_set_secondary (PluginHandle * plugin)
     if (sop && ! sop->init ())
         sop = nullptr;
 
-    if (s_input && s_output && sop && setup_plugin (sop, FMT_FLOAT))
-        s_secondary = true;
+    if (s_input && aud_get_bool (0, "record"))
+        setup_secondary (false);
 
     UNLOCK_MINOR;
     return (! plugin || sop);
 }
+
+static void record_settings_changed (void *, void *)
+{
+    LOCK_MINOR;
+
+    if (s_input && aud_get_bool (0, "record"))
+        setup_secondary (false);
+    else
+        cleanup_secondary ();
+
+    UNLOCK_MINOR;
+}
+
+void output_init ()
+{
+    hook_associate ("set record", record_settings_changed, nullptr);
+    hook_associate ("set record_stream", record_settings_changed, nullptr);
+}
+
+void output_cleanup ()
+{
+    hook_dissociate ("set record", record_settings_changed);
+    hook_dissociate ("set record_stream", record_settings_changed);
+}
