@@ -39,8 +39,8 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
     QBoxLayout * parent_layout = nullptr;
     QBoxLayout * parent_orig_layout = nullptr;
 
-    BooleanWidget * parent_widget = nullptr;
-    QButtonGroup * radio_btn_group = nullptr;
+    ParentWidget * parent_widget = nullptr;
+    QButtonGroup * radio_btn_group[2] = {nullptr, nullptr};
 
     for (const PreferencesWidget & w : widgets)
     {
@@ -54,14 +54,13 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
 
                 /* create new layout for child widgets */
                 if (dynamic_cast<QHBoxLayout *> (parent_layout))
-                    layout = new QHBoxLayout;
+                    layout = make_hbox (nullptr, sizes.TwoPt);
                 else
                 {
-                    layout = new QVBoxLayout;
-                    layout->setContentsMargins (12, 0, 0, 0);
+                    layout = make_vbox (nullptr, sizes.TwoPt);
+                    layout->setContentsMargins (sizes.EightPt, 0, 0, 0);
                 }
 
-                layout->setSpacing (parent_layout->spacing ());
                 parent_layout->addLayout (layout);
 
                 orig_layout = layout;
@@ -84,13 +83,15 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
 
             /* enable/disable child widgets */
             if (parent_widget)
-                parent_widget->update ();
+                parent_widget->update_from_cfg ();
 
             parent_widget = nullptr;
         }
 
-        if (radio_btn_group && w.type != PreferencesWidget::RadioButton)
-            radio_btn_group = nullptr;
+        if (w.type != PreferencesWidget::RadioButton)
+            radio_btn_group[w.child] = nullptr;
+        if (! w.child)
+            radio_btn_group[true] = nullptr;
 
         switch (w.type)
         {
@@ -115,16 +116,15 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
 
             if (strstr (w.label, "<b>"))
             {
-                /* double spacing above a header */
+                /* extra spacing above a header */
                 if (orig_layout->itemAt (0))
-                    orig_layout->addSpacing (orig_layout->spacing ());
+                    orig_layout->addSpacing (sizes.EightPt);
 
                 orig_layout->addWidget (label);
 
                 /* create indented layout below header */
-                layout = new QVBoxLayout;
-                layout->setContentsMargins (12, 0, 0, 0);
-                layout->setSpacing (orig_layout->spacing ());
+                layout = make_vbox (nullptr, sizes.TwoPt);
+                layout->setContentsMargins (sizes.EightPt, 0, 0, 0);
                 orig_layout->addLayout (layout);
             }
             else
@@ -149,15 +149,24 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
             break;
 
         case PreferencesWidget::Entry:
+        /* TODO: implement file chooser */
+        case PreferencesWidget::FileEntry:
             layout->addWidget (new StringWidget (& w, domain));
             break;
 
         case PreferencesWidget::RadioButton:
-            if (! radio_btn_group)
-                radio_btn_group = new QButtonGroup;
+        {
+            if (! radio_btn_group[w.child])
+                radio_btn_group[w.child] = new QButtonGroup;
+
+            auto radio_btn = new RadioButtonWidget (& w, domain, radio_btn_group[w.child]);
+            layout->addWidget (radio_btn);
+
+            if (! w.child)
+                parent_widget = radio_btn;
 
-            layout->addWidget (new RadioButtonWidget (& w, domain, radio_btn_group));
             break;
+        }
 
         case PreferencesWidget::FontButton:
             /* XXX: unimplemented */
@@ -191,7 +200,12 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
         {
             QFrame * f = new QFrame;
             f->setFrameShape (w.data.separator.horizontal ? QFrame::HLine : QFrame::VLine);
+            f->setFrameShadow (QFrame::Sunken);
+
+            layout->addSpacing (sizes.FourPt);
             layout->addWidget (f);
+            layout->addSpacing (sizes.FourPt);
+
             break;
         }
 
@@ -204,7 +218,7 @@ void prefs_populate (QBoxLayout * layout, ArrayRef<PreferencesWidget> widgets, c
 
     /* enable/disable child widgets */
     if (parent_widget)
-        parent_widget->update ();
+        parent_widget->update_from_cfg ();
 }
 
 } // namespace audqt
